#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>

// We will compute pi using monte carlo method.
// 1) pick large number of random x and y coordinates
// 2) root will dirstribute the random points over all the tasks
// 3) every task computes home many of its points are inside the circle
//    point is inside circle if sqrt(x^2+y%2) <= 1
// 4) every tasks sends back to root its local @inside
// 5) roo will receive all the partial results and add them up
// 6) root will compute pi using the formaly pi = inside*4 / total
// 7 all tasks will finish

// this function takes a input one array, containing x and y coordinates
// and returns how many points are inside the circle using sqrt(x^2+y^2) <= 1
int count_inside(float array[], int size) {
  int total_inside=0;
  int half = size/2;
  for ( int i=0;i<half;++i) {
    float x = array[i];
    float y = array[half+i];
    if ((sqrt(x*x + y*y)) <= 1) {
      ++total_inside;
    }
  }
  return total_inside;
}


// this function takes an array and fills it with random numbers 
// between 0 and 1
void fill_random(float array[], int size) {
  for (int i = 0; i< size;++i) {
    array[i]=(float)rand() / (float)RAND_MAX ;
  }
}

int main(int argc, char **argv){
    int number, size, rank;
    MPI_Status status;
    double t1, t2, elapsed,t3,t4;
    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    if (size < 2){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0){
      printf("Enter number of random points:");
      scanf("%d", &number);
      t1 = MPI_Wtime();
      int total_xy = (number/size)*2;
      
      // will create an array to store random x,y
      // Instead of creating an x and y array, will just create a single
      // array of double size that holds both x and y
      // and instead of creating one huge array of size number*2 will create
      // array with size (number/size)*2 which can store the random points for a single task
      // and re-use the array when sending data to every task
      float*  random_points = (float*)calloc(total_xy, sizeof(float));


      // the root (rank 0) has to sent the data ( random points) to the other tasks.
      // Will use a loop for this
      for (i=1; i<size; i++) {
         // TODO distribute the data over all the tasks
      }

      // root will also compute part of the results
      fill_random(random_points,total_xy);
      int local_inside=count_inside(random_points,total_xy);
     
      // get the #inside from all the other tasks
      for (i=1; i<size; i++) {
	int remote_inside;
	// wait to receive the next value, all tasks will send their results
	// combine the results
      }

      // finally root will compute pi
      float mc_pi = 4.0*((float)local_inside/number);
      t2 = MPI_Wtime();
      printf("Value of Pi: %.10f, total time: %.10f\n",mc_pi,(t2-t1));
    }else{

      // Other task will receive a number of random points and compute how many of those points are inside circel
      // TODO receive the information
      int local_inside=count_inside(random_points,total_xy);

      // TODO send results back to root
      
    }  

    MPI_Finalize();
}
