#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>

// We will compute pi using monte carlo method.
// 1) pick large number of random x and y coordinates
// 2) root will broadcast total count
// 3) every task creates array x,y of random points and computes 
//    home many of its points are inside the circle
//    point is inside circle if sqrt(x^2+y%2) <= 1
// 4) every tasks sends back to root its local @inside
// 5) roo will receive all the partial results and add them up
// 6) root will compute pi using the formaly pi = inside*4 / total
// 7 all tasks will finish

int count_inside(float array[], int size) {
  int total_inside=0;
  int half = size/2;
  for ( int i=0;i<half;++i) {
    float x = array[i];
    float y = array[half+i];
    if ((sqrt(x*x + y*y)) <= 1) {
      ++total_inside;
    }
  }
  return total_inside;
}

void fill_random(float array[], int size) {
  for (int i = 0; i< size;++i) {
    // todo call the random function
    array[i]=(float)rand() / (float)RAND_MAX ;
  }
}

int main(int argc, char **argv){
    int number, size, rank;
    MPI_Status status;
    double t1, t2, elapsed,t3,t4;
    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    if (size < 2){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0){
      printf("Enter number of random points:");
      scanf("%d", &number);
    }

    t1 = MPI_Wtime();
    MPI_Bcast(&number, 1, MPI_INT, 0, MPI_COMM_WORLD);
    int total_points = number/size;
    
    // will create an array to store random x,y
    // Instead of creating an x and y array, will just create a single
    // array of double size that holds both x and y
    int total_xy = total_points*2;
    float*  random_points = (float*)calloc(total_xy, sizeof(float));
    fill_random(random_points,total_xy);
    int local_inside=count_inside(random_points,total_xy);

    // all tasks did their computation, now time to get the results
    // task 0 will receive all the partial results, all other tasks
    // will send their results to task 0

    // TODO instead of gather do a reduction to get all the partial results

    if (rank == 0) {
      // finally root will compute pi
      float mc_pi = 4.0*((float)total_inside/number);
      t2 = MPI_Wtime();
      printf("Value of Pi: %.10f, total time: %.10f\n",mc_pi,(t2-t1));
    } 

    MPI_Finalize();
}
