#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>

// We will compute pi using monte carlo method.
// 1) pick large number of random x and y coordinates
// 2) root will dirstribute the random points over all the tasks
// 3) every task computes home many of its points are inside the circle
//    point is inside circle if sqrt(x^2+y%2) <= 1
// 4) every tasks sends back to root its local @inside
// 5) roo will receive all the partial results and add them up
// 6) root will compute pi using the formaly pi = inside*4 / total
// 7 all tasks will finish

int count_inside(float array[], int size) {
  int total_inside=0;
  int half = size/2;
  for ( int i=0;i<half;++i) {
    float x = array[i];
    float y = array[half+i];
    if ((sqrt(x*x + y*y)) <= 1) {
      ++total_inside;
    }
  }
  return total_inside;
}

void fill_random(float array[], int size) {
  for (int i = 0; i< size;++i) {
    // todo call the random function
    array[i]=(float)rand() / (float)RAND_MAX ;
  }
}

int main(int argc, char **argv){
    int number, size, rank;
    MPI_Status status;
    double t1, t2, elapsed,t3,t4;
    int i;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    if (size < 2){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0){
      printf("Enter number of random points:");
      scanf("%d", &number);
      //number=1000000000;
      t1 = MPI_Wtime();
      int total_xy = (number/size)*2;
      
      // will create an array to store random x,y
      // Instead of creating an x and y array, will just create a single
      // array of double size that holds both x and y
      float*  random_points = (float*)calloc(total_xy, sizeof(float));

      for (i=1; i<size; i++) {
	// fill the array with random values between 0 and 1
	fill_random(random_points,total_xy);

	// send the number of x+y coordinates 
	MPI_Send(&total_xy, 1, MPI_INT, i, 0, MPI_COMM_WORLD);

	// send the actual x,y coordinates
	MPI_Send(random_points,total_xy,MPI_REAL,i,0,MPI_COMM_WORLD);
      }

      // root will also compute part of the results
      fill_random(random_points,total_xy);

      t3 = MPI_Wtime();
      int local_inside=count_inside(random_points,total_xy);
      t4 = MPI_Wtime();
      printf("time root to calculate #count : %.10f\n",(t4-t3));
      // get the #inside from all the other tasks
      for (i=1; i<size; i++) {
	int remote_inside;
	// wait to receive the next value, all tasks will send their results
	MPI_Recv(&remote_inside, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
	local_inside = local_inside + remote_inside;
      }

      // finally root will compute pi
      float mc_pi = 4.0*((float)local_inside/number);
      t2 = MPI_Wtime();
      printf("Value of Pi: %.10f, total time: %.10f\n",mc_pi,(t2-t1));
    }else{

      // task will receive number of x+y and create buffer to store them
      int total_xy;
      MPI_Recv(&total_xy, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

      float*  random_points = (float *)calloc(total_xy, sizeof(float));

      // receive the actual  x,y coordinates
      MPI_Recv(random_points, total_xy, MPI_REAL, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      int local_inside=count_inside(random_points,total_xy);

      // send results back to root
      MPI_Send(&local_inside, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
      
    }  

    t1 = MPI_Wtime();

    MPI_Finalize();
}
