#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>

// We will compute pi using monte carlo method.
// 1) pick large number of random x and y coordinates
// 2) root will scatter the random points over all the tasks
// 3) every task computes home many of its points are inside the circle
//    point is inside circle if sqrt(x^2+y%2) <= 1
// 4) all partial results will be gathered bu root
// 6) root will compute pi using the formaly pi = inside*4 / total
// 7 all tasks will finish

int count_inside(float array[], int size) {
  int total_inside=0;
  int half = size/2;
  for ( int i=0;i<half;++i) {
    float x = array[i];
    float y = array[half+i];
    if ((sqrt(x*x + y*y)) <= 1) {
      ++total_inside;
    }
  }
  return total_inside;
}

void fill_random(float array[], int size) {
  for (int i = 0; i< size;++i) {
    // todo call the random function
    array[i]=(float)rand() / (float)RAND_MAX ;
  }
}

int main(int argc, char **argv){
    int number, size, rank;
    MPI_Status status;
    double t1, t2;
    float *all_random_points, *random_points;
    int i, total, total_xy;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    if (size < 2){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0){
      printf("Enter number of random points:");
      scanf("%d", &number);
      //number=1000000000;
      t1 = MPI_Wtime();
    }

   
    MPI_Bcast(&number, 1, MPI_INT, 0, MPI_COMM_WORLD);

    int total_points = number/size;
    // will create an array to store random x,y
    // Instead of creating an x and y array, will just create a single
    // array of double size that holds both x and y
    total_xy = total_points*2;
    random_points = (float*)calloc(total_xy, sizeof(float));
    fill_random(random_points,total_xy);

    // every tasks computes the local #points inside
    int local_inside=count_inside(random_points,total_xy);

    // root gathers the partial results
    int* total_inside = (int*)calloc(size, sizeof(int));
    MPI_Gather(&local_inside, 1, MPI_INT, total_inside, 1, MPI_INT, 0, MPI_COMM_WORLD);

    if (rank == 0) {
      // finally root will compute pi
      int sum =0;
      for (int i=0;i<size;++i) {
	sum=sum+total_inside[i];
	printf("next local_inside: %d\n",total_inside[i]);
      }
      float mc_pi = 4.0*((float)sum/number);
      t2 = MPI_Wtime();
      printf("Value of Pi: %.10f, total time: %.10f\n",mc_pi,(t2-t1));

    } 

    MPI_Finalize();
}
