#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>

// We will compute pi using monte carlo method.
// 1) pick large number of random x and y coordinates
// 2) root will scatter the random points over all the tasks
// 3) every task computes home many of its points are inside the circle
//    point is inside circle if sqrt(x^2+y%2) <= 1
// 4) every tasks sends back to root its local @inside
// 5) roo will receive all the partial results and add them up
// 6) root will compute pi using the formaly pi = inside*4 / total
// 7 all tasks will finish

int count_inside(float array[], int size) {
  int total_inside=0;
  int half = size/2;
  for ( int i=0;i<half;++i) {
    float x = array[i];
    float y = array[half+i];
    if ((sqrt(x*x + y*y)) <= 1) {
      ++total_inside;
    }
  }
  return total_inside;
}

void fill_random(float array[], int size) {
  for (int i = 0; i< size;++i) {
    // todo call the random function
    array[i]=(float)rand() / (float)RAND_MAX ;
  }
}

int main(int argc, char **argv){
    int number, size, rank;
    MPI_Status status;
    double t1, t2;
    float *all_random_points, *random_points;
    int i, total, total_xy;

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    if (size < 2){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == 0){
      printf("Enter number of random points:");
      scanf("%d", &number);
      //number=1000000000;
      t1 = MPI_Wtime();

      // create an array to store random x,y
      // Instead of creating an x and y array, will just create a single
      // array of double size that holds both x and y
      total = 2*number;
      all_random_points = (float*)calloc(total, sizeof(float));
      fill_random(all_random_points,total);

      // reserve space for the subset
      total_xy = total/size;
      random_points = (float*)calloc(total_xy, sizeof(float));

      for (int i=1; i<size; i++) {
	// send the number of x+y coordinates 
	MPI_Send(&total_xy, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
      }

    } else {
      MPI_Recv(&total_xy, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
      // reserve space to store the actual numbers
      random_points = (float*)calloc(total_xy, sizeof(float));
    }


    // root scatters the random points, all tasks will receive a subset
    MPI_Scatter(all_random_points, total_xy, MPI_REAL, random_points, total_xy, MPI_REAL, 0, MPI_COMM_WORLD);
   
    // every task will compute number of points inside
    int local_inside=count_inside(random_points,total_xy);

   
    if (rank == 0) {
      for (int i=1; i<size; i++) {
        int remote_inside;
        // wait to receive the next value, all tasks will send their results
        MPI_Recv(&remote_inside, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
        local_inside = local_inside + remote_inside;
      }

      // finally root will compute pi
      float mc_pi = 4.0*((float)local_inside/number);
      t2 = MPI_Wtime();
      printf("Value of Pi: %.10f, total time: %.10f\n",mc_pi,(t2-t1));

    } else {
      MPI_Send(&local_inside, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
    }

    MPI_Finalize();
}
