/* calculate PI using 
    /1
    |   4.0
    | -------- dx = PI
    | (1+x*x)
    /0
*/

#include <stdio.h>
#include <mpi.h>

#define N 10000
#define BLK 4
#define MIN(x,y)   ((x)<(y)?(x):(y))

int main(int argc, char **argv){

    double h, x, sum, pi;
    int    myid, nprocs, i,j;
    MPI_Request *reqs;

    double t1, t2;
    

    MPI_Init(&argc, &argv);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (nprocs > 100){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);

    t1 = MPI_Wtime();

    h = 1.0/(double)N;
    sum = 0.0;

    // block cyclic distribution 
    for (i=myid*BLK+1; i<=N; i+=nprocs*BLK){
       for (j=i; j<=MIN(N,i+BLK-1); j++){
           x = h*(j-0.5);
           sum += 4.0/(1.0+x*x); 
       }
    }
    sum = sum*h;

    // call MPI_Reduce to get the value of sum (PI) in process 0
    MPI_Reduce(&sum, &pi, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    t2 = MPI_Wtime();

    if (myid == 0){
        printf("Pi is approx. %.10f\n", pi);
        printf("Total running time is %10.8f\n", t2-t1);
    }

    MPI_Finalize();

}
