/* calculate PI using 
    /1
    |   4.0
    | -------- dx = PI
    | (1+x*x)
    /0
*/

#include <stdio.h>
#include <mpi.h>
#include <omp.h>

#define N 100000000
#define MIN(x,y)   ((x)<(y)?(x):(y))

void block_map(int n1, int n2, int nprocs, int myid, int *l1, int *l2){
    int block, rem;

    block = (n2-n1+1)/nprocs;
    rem   = (n2-n1+1)%nprocs;

    if (myid < rem){
        block++;
        *l1 = n1+myid*block;
    }else
        *l1 = n1+rem+block*myid;
    *l2 = *l1+block-1;
}

int main(int argc, char **argv){

    double h, x, sum, pi;
    int    myid, nprocs, i,j;
    int    l1, l2;
    MPI_Request *reqs;
    int    provided;

    double t1, t2;
    

//  MPI_Init_thread(&argc, &argv, MPI_THREAD_SINGLE, &provided);
//  MPI_Init_thread(&argc, &argv, MPI_THREAD_FUNNELED, &provided);
//  MPI_Init_thread(&argc, &argv, MPI_THREAD_SERIALIZED, &provided);
    MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &provided);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
    if (nprocs > 100){
        MPI_Abort(MPI_COMM_WORLD, 99);
    }
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);

    t1 = MPI_Wtime();

    h = 1.0/(double)N;
    sum = 0.0;

    // block distribution
    block_map(1,N,nprocs,myid,&l1,&l2);
#pragma omp parallel for private(i, x) reduction(+: sum)
    for (i=l1; i<=l2; i++){
       x = h*(i-0.5);
       sum += 4.0/(1.0+x*x); 
    }
    sum = sum*h;

        // call MPI_Reduce to get the value of sum (PI) in process 0
    MPI_Reduce(&sum, &pi, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    t2 = MPI_Wtime();

    if (myid == 0){
        printf("Pi is approx. %.20f\n", pi);
        printf("Total running time is %10.8f\n", t2-t1);
    }

    MPI_Finalize();

}
