! calculate PI using 
!   /1
!   |   4.0
!   | -------- dx = PI
!   | (1+x*x)
!   /0
!

program ex5
use mpi
implicit none

integer,parameter::N=20000, BLK=4, DP=selected_real_kind(15)

real(kind=DP)::h, x, sum, pi
integer        myid, nprocs, i,j
integer        l1, l2, ierr

real(kind=DP)::t1, t2

    
call MPI_Init(ierr)
call MPI_Comm_size(MPI_COMM_WORLD, nprocs,ierr)
    
if (nprocs > 100) then
    call MPI_Abort(MPI_COMM_WORLD, 99, ierr)
endif
    
call MPI_Comm_rank(MPI_COMM_WORLD, myid, ierr)

t1 = MPI_Wtime()

h = 1.0d0/REAL(N,KIND=DP)
sum = 0.0d0

!block distribution
call block_map(1,N,nprocs,myid,l1,l2)
do i=l1, l2
   x = h*(i-0.5d0)
    sum = sum + 4.0d0/(1.0d0+x*x) 
enddo
sum = sum*h


call MPI_REDUCE(sum, pi, 1, MPI_REAL8, MPI_SUM, 0, MPI_COMM_WORLD, ierr)

t2 = MPI_Wtime()

if (myid .eq. 0) then
    print *, "Pi is approx. ", pi
    print *, "Total running time is ", t2 - t1
endif

call MPI_FINALIZE(ierr)

end program ex5



subroutine block_map(n1, n2, nprocs, myid, l1, l2)
integer n1, n2, nprocs, myid, l1, l2

integer block, rem

block =(n2-n1+1)/nprocs
rem = mod(n2-n1+1,nprocs)
if (myid < rem) then
    block = block+1
    l1 = n1+myid*block
else
    l1 = n1+rem+block*myid
endif
l2 = l1+block-1

end subroutine block_map
